# -*- coding: utf-8 -*-

"""
God object that controls API communication.
"""

import os.path
import requests
import threading
from . import util
from . import aggregation


class SDK(object):

    """
    Consumer for the Geezeo API.

    The use of this package starts with an :class:`SDK` instance, which is
    created with the authentication information required by Geezeo's web API.

    The constructor can raise an :class:`UnauthorizedError` if ``api_key`` is
    incorrect, or a :class:`DoesNotExistError` if the ``user_id`` doesn't
    correspond to an actual user.

    All methods can raise :class:`ServerError` if an error occurred on the
    server, or :class:`NetworkError` if something goes wrong while attempting
    to communicate with the server.

    .. attribute:: api_key

        Authentication key for the partner institution using this SDK. If
        ``api_key`` is invalid, the constructor will raise an
        :class:`UnauthorizedError`.

    .. attribute:: user

        :class:`User` on whose behalf the SDK is being used. Note that the
        :class:`SDK` object is constructed with the string ``user_id``,
        and this attribute will be populated with the corresponding
        :class:`User` during construction.

    .. attribute:: url

        URL for the partner institution's web site. This will be the root of
        URLs for all requests made to the API.
    """

    def __init__(self, api_key, user_id, url):
        self.api_key = api_key
        self.url = url
        self.session = requests.Session()

        # This will raise UnauthorizedError if not authorized,
        # NotFoundError if the user does not exist,
        # NetworkError if the url is invalid.
        self.user = self.get_user(user_id)

        self.pending = threading.Thread(target=self.clear_pending_accounts)
        self.pending.start()

    def request(self, method, path, **kwargs):
        """
        Make a request of the API.

        Additional arguments in ``kwargs`` will be passed to an underlying
        :func:`requests.request` call.

        .. _Geezeo: http://developers.geezeo.com

        :param str method: ``'GET'``, ``'POST'``, etc.
        :param str path: portion of the URL that follows :attr:`self.url
            <url>`. This should match the HTTP path in the `Geezeo`_ docs,
             e.g. ``'/api/v2/ping'``. Note that GET parameters should
            be managed with the ``params`` argument instead.
        :return: dictionary parsed from the JSON returned by the endpoint.
        """

        # s/get/GET/
        method = method.upper()

        # If the path starts with a slash, it is interpreted as a path from
        # root (consequence of using an os.path function to join components).
        if path.startswith('/'):
            path = path[1:]
        url = os.path.join(self.url, path)

        # Requests use HTTP basic auth, with the api_key as the username and
        # no password.
        auth = (self.api_key, '')

        try:
            request = requests.Request(method, url, auth=auth, **kwargs)
            content_type = 'application/x-www-form-urlencoded'
            request.headers['Content-Type'] = content_type
            prepared_request = request.prepare()
            response = self.session.send(prepared_request)
        except requests.exceptions.RequestException as e:
            # The requests.request function raised an exception, indicating
            # that we didn't communicate successfully with the server.
            raise util.NetworkError(e)
        else:
            # We got a response, but not necessarily a successful one.
            if response.status_code == 200:
                return response.json()
            elif response.status_code == 401:
                raise util.UnauthorizedError(response)
            elif response.status_code == 404:
                raise util.DoesNotExistError(response)
            elif response.status_code == 405:
                msg = 'invalid method for {0}: {1}'.format(path, method)
                raise ValueError(msg)
            elif response.status_code == 500:
                raise util.ServerError(response)

    def ping(self):
        """
        Raise :class:`UnauthorizedError` if the SDK was not initialized
        correctly.
        """

        self.request('GET', '/api/v2/ping')

    def get_user(self, user_id):
        """
        Fetch the User with the user_id provided.

        :param string user_id: unique identifier, corresponds to
            :attr:`User.id`
        :raises DoesNotExistError: if no user with ID ``user_id`` exists.
        :return: the :class:`User` in question
        """

        body = self.request('GET', '/api/v2/users/{0}'.format(user_id))
        return util.User(body['users'][0])

    def pending_account_ids(self):
        """
        Generate IDs for pending accounts, which must be destroyed!
        """

        url = '/api/v2/users/{0}/pending_accounts'.format(self.user.id)
        data = self.request('GET', url)
        for fi in data['pending_accounts']:
            for id in fi['account_ids']:
                yield id

    def clear_pending_accounts(self):
        """
        Delete all of the user's pending accounts.

        This function retrieves a list of the account IDs for all of the
        user's pending accounts. It then spawns a thread for each one, which
        uses the API to delete that account. The function does not return
        until all threads have finished running.
        """

        # Given a pending account ID, make a request to delete it.
        def delete_account(id):
            url_template = '/api/v2/users/{0}/pending_accounts/{1}'
            self.request('DELETE', url_template.format(self.user.id, id))

        # Spawn a thread to delete each pending account.
        delete_threads = []
        for id in self.pending_account_ids():
            thread = threading.Thread(target=delete_account, args=(id,))
            delete_threads.append(thread)
            thread.start()

        # Wait for all pending accounts to be deleted.
        for thread in delete_threads:
            thread.join()

    def search_institutions(self, search_string, scope=None, page=1):
        """
        Search the set of institutions aggregated by the aggregation provider.

        .. note::

            If the search string is sufficiently vague, the results will be
            truncated. Since the API backing this function is paginated, the
            truncation will be to three pages of results.

        :param str search_string: keywords to search for
        :param str scope: if present, must be ``'name'`` or ``'url'``.
            determines how the institution needs to match ``search_string``.
        :param int page: (optional) page of results to retrieve. Get the
            first page by default.
        :return: :class:`PagedResults` whose :attr:`~PagedResults.data`
            is a list of :class:`AuthPrompt` objects for institutions
            matching the ``search_string``.
        """

        assert isinstance(page, int), 'page must be an integer'
        assert page > 0, 'page must be positive'
        params = {
            'q': search_string,
            'page': page,
        }

        if scope is not None:
            assert scope in ('name', 'url'), "Scope must be 'name' or 'url'"
            params['scope'] = scope

        data = self.request('GET', '/api/v2/ce_fis/search', params=params)
        prompts = [aggregation.AuthPrompt(item) for item in data['ce_fis']]
        current_page = data['meta']['current_page']
        last_page = data['meta']['total_pages']
        return util.PagedResults(prompts, current_page, last_page)

    def get_all_institutions(self, page=1):
        """
        Return a page from the set of all institutions.

        :param int page: (optional) page of results to retrieve. Get the
            first page by default.
        :return: :class:`PagedResults` whose :attr:`~PagedResults.data` is a
            list of :class:`AuthPrompt` objects.
        """

        assert isinstance(page, int), 'page must be an integer'
        assert page > 0, 'page must be positive'
        params = {
            'page': page
        }

        data = self.request('GET', '/api/v2/ce_fis', params=params)
        prompts = [aggregation.AuthPrompt(item) for item in data['ce_fis']]
        current_page = data['meta']['current_page']
        last_page = data['meta']['total_pages']
        return util.PagedResults(prompts, current_page, last_page)

    def get_featured_institutions(self):
        """
        Return the list of featured institutions for the current partner.

        :return: list of :class:`AuthPrompt` objects for institutions featured
            by the current partner.
        """

        # This is not available in the current API. Return any old FIs.
        return self.get_all_institutions(page=1).data

    def authenticate(self, submit_key, parameters):
        """
        Send login information to the aggregation provider for authentication.

        :param str submit_key: the :attr:`submit_key <AuthPrompt.submit_key>`
            attribute of the prompt for which to submit auth details
        :param dict parameters: dictionary (or dictionary-like object) mapping
            the IDs in this form's :attr:`parameters <AuthPrompt.parameters>`
            attribute to the user's corsubmiting login information
        :raises MFARequiredError: if the institution requires multi-factor
            authentication
        :raises UnauthorizedError: if the authentication fails
        :return: the :class:`AggregatedInstitution` for which the user has
            authenticated
        """

        self.pending.join()
        submit_data = aggregation.decode_submit_key(submit_key)

        # Construct and send the request.

        if submit_data['type'] == 'auth':
            method = 'POST'
            url = '/api/v2/users/{0}/ce_fis'.format(self.user.id)
            fields = {
                'id': submit_data['id'],
            }
            for key, value in parameters.items():
                fields['credentials[login_params][{0}]'.format(key)] = value

        elif submit_data['type'] == 'mfa':
            method = 'PUT'
            url = '/api/v2/users/{0}/ce_fis/{1}'.format(self.user.id,
                                                        submit_data['id'])
            fields = {
                'harvest_id': submit_data['harvest_id'],
                'login_id': submit_data['login_id'],
                'session_key': submit_data['session_key'],
            }
            for key, value in parameters.items():
                fields['mfa_responses[{0}]'.format(key)] = value

        response = self.request(method, url, data=fields)

        # Determine what kind of information we got back, based on the JSON
        # data from the server.

        # Success, accompanied by a list of accounts.
        if 'accounts' in response:
            accounts = []
            classifications = {}
            for item in response['accounts']:

                # Prepare the list of accounts for the AggregatedInstitution
                # return value.
                accounts.append(util.Account(item))

                # Track the account type and ext type of each account for
                # classification.
                id = item['id']
                try:
                    acct_type = item['cashedge_account_type']['acct_type']
                    ext_type = item['cashedge_account_type']['ext_type']
                    types = (acct_type, ext_type)
                except (KeyError, TypeError):
                    types = None
                classifications[id] = types

            self.classify_accounts(classifications)
            return aggregation.AggregatedInstitution(submit_data, accounts)

        # MFA challenge
        elif 'response' in response:
            message = response['response']['message']
            data = response['response']['data']
            if 'mfa_parameters' in data:
                raise aggregation.MFARequiredError(message, data, submit_data)
            elif data.get('code') == 'bad_credentials':
                raise util.UnauthorizedError(response)

    def update_authentication(self, submit_key, parameters):
        """
        :param str submit_key: :attr:`~AggregatedInstitution.submit_key` of an
            institution whose auth info will be updated
        :param dict parameters: dictionary mapping the :attr:`key
            <AuthParameter.key>` attributes from :attr:`parameters
            <AggregatedInstitution.parameters>` to the new values for each one
        :raises MFARequiredError: if the institution requires multi-factor
            auth to change parameters
        :raises TypeError: if the parameters provided don't match the prompt
        """

        self.pending.join()
        submit_data = aggregation.decode_submit_key(submit_key)
        institution_id = submit_data['id']

        # The old CashEdge API, for reasons that surpass understanding,
        # requires an account ID to update the authentication for an
        # institution. To preserve the sanity of our semantics, we use
        # the institution ID to look up an account ID for any
        # corresponding account, then use that account ID.
        #
        # This is wildly inefficient; we are resting on two things:
        #   1. Our plan to make this implementation obsolete.
        #   2. Our standing disclaimer that this call will take a long time.

        account_id = None
        accounts_url = '/api/v2/users/{0}/accounts'.format(self.user.id)
        accounts_response = self.request('GET', accounts_url)

        # Iterate through all the user's accounts until we find one from the
        # institution whose auth info we wish to update.

        for account in accounts_response['accounts']:
            if str(account['fi']['id']) == institution_id:
                account_id = account['id']
                break

        if account_id is None:
            raise util.DoesNotExistError('No accounts usable for updating '
                                         'credentials.')

        url_template = '/api/v2/users/{0}/accounts/{1}/update_credentials'
        url = url_template.format(self.user.id, account_id)

        # Populate the form data from the new values provided.

        if submit_data['type'] == 'institution':
            fields = {}
            for key, value in parameters:
                fields['credentials[login_params][{0}]'.format(key)] = value
        else:
            fields = {
                'harvest_id': submit_data['harvest_id'],
                'login_id': submit_data['login_id'],
                'session_key': submit_data['session_key'],
            }
            for key, value in parameters:
                fields['mfa_responses[{0}]'.format(key)] = value

        response = self.request('PUT', url, data=fields)
        if 'response' in response:
            message = response['response']
            data = response['data']
            raise aggregation.MFARequiredError(message, data, submit_data)

    def classify_accounts(self, classifications):
        """
        Update the account types for some set of accounts.

        :param dict classifications: Dictionary mapping string account IDs to
            either an ``(account_type, ext_type)`` pair or ``None``. If the
            value is ``None``, the account will be dropped from aggregation.
        """

        fields = {}
        for account_id, types in classifications.items():
            key = 'accounts[{0}][type_code]'.format(account_id)
            if types is None:
                value = 'ignore'
            else:
                account_type, ext_type = types
                value = '{0},{1}'.format(account_type, ext_type)
            fields[key] = value
        url = '/api/v2/users/{0}/accounts/classify'.format(self.user.id)
        response = self.request('PUT', url, data=fields)
        assert 'accounts' in response, 'Unexpected classification response.'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_geezeo
----------------------------------

Tests for `geezeo` module.
"""

import unittest
import geezeo
import six
from pytest import raises


class TestSDK(unittest.TestCase):

    url = 'https://geez3o.geezeo.com'
    user_id = 'david-donna-amston'
    api_key = ('b695f27def0277d34c726576772c5223b84bbd43e23391c5da2f344f5'
               '3c3e1a41cbcfe93104d010790d84c376ca25863a502039c6fe9c034da'
               '73f6edb7cd02e9')

    def test_successful_sdk_creation(self):
        geezeo.SDK(self.api_key, self.user_id, self.url)

    def test_url_is_set_correctly(self):
        sdk = geezeo.SDK(self.api_key, self.user_id, self.url)
        assert sdk.url == self.url

    def test_user_is_set_correctly(self):
        "Make sure we get a User with the correct user_id."
        sdk = geezeo.SDK(self.api_key, self.user_id, self.url)
        assert isinstance(sdk.user, geezeo.User)
        assert sdk.user.id, self.user_id

    def test_unsuccessful_sdk_creation_with_bad_api_key(self):
        "Use a mangled key, and expect failure."
        with raises(geezeo.UnauthorizedError):
            geezeo.SDK(self.api_key[10:], self.user_id, self.url)

    def test_unsuccessful_sdk_creation_with_bad_user_id(self):
        "Use a mangled key, and expect failure."
        with raises(geezeo.DoesNotExistError):
            geezeo.SDK(self.api_key, self.user_id[3:], self.url)

    def test_unsuccessful_sdk_creation_with_bad_url(self):
        "Use a mangled key, and expect failure."
        with raises(geezeo.NetworkError):
            geezeo.SDK(self.api_key, self.user_id, self.url + '.horse')

    def test_unsuccessful_sdk_creation_with_bad_existing_url(self):
        "Use a mangled key, and expect failure."
        with raises(geezeo.DoesNotExistError):
            geezeo.SDK(self.api_key, self.user_id, 'https://www.google.com')


class TestAggregation(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.sdk = geezeo.SDK(TestSDK.api_key, TestSDK.user_id, TestSDK.url)

    def test_get_all_fails_with_non_integer_page(self):
        with raises(Exception):
            self.sdk.get_all_institutions(page='1')

    def test_get_all_fails_with_zero_page(self):
        with raises(Exception):
            self.sdk.get_all_institutions(page=0)

    def test_get_all_fails_with_negative_one_page(self):
        with raises(Exception):
            self.sdk.get_all_institutions(page=-1)

    def test_get_all_is_empty_with_absurd_page(self):
        results = self.sdk.get_all_institutions(page=1000000)
        assert isinstance(results, geezeo.PagedResults)
        assert results.data == []

    def test_get_all_returns_paged_results(self):
        results = self.sdk.get_all_institutions()
        assert isinstance(results, geezeo.PagedResults)
        assert len(results) > 0
        assert all(isinstance(p, geezeo.AuthPrompt) for p in results)

    def test_get_all_returns_different_results_for_different_pages(self):
        r1 = self.sdk.get_all_institutions(page=1)
        r2 = self.sdk.get_all_institutions(page=2)
        assert r1.data != r2.data
        assert r1.current_page == 1
        assert r2.current_page == 2

    def test_get_featured_institutions_returns_prompt_list(self):
        prompts = self.sdk.get_featured_institutions()
        assert isinstance(prompts, list)
        assert len(prompts) > 1
        assert all(isinstance(p, geezeo.AuthPrompt) for p in prompts)

    def test_gibberish_search_string_yields_no_results(self):
        result = self.sdk.search_institutions('alkdfoqweifyow764ryoaiu')
        assert isinstance(result, geezeo.PagedResults)
        assert len(result) == 0

    def test_bad_search_scope_raises_exception(self):
        for bad_scope in [1, 'foo', ' url ']:
            with raises(Exception):
                self.sdk.search_institutions('bank', scope=bad_scope)

    def test_bad_search_page_raises_exception(self):
        for bad_page in [-1, 0, 1.2, None]:
            with raises(Exception):
                self.sdk.search_institutions('bank', page=bad_page)

    def test_high_search_page_yields_no_results(self):
        results = self.sdk.search_institutions('bank', page=1000000)
        assert isinstance(results, geezeo.PagedResults)
        assert results.data == []

    def test_search_scope_can_be_absent(self):
        self.sdk.search_institutions('bank')

    def test_search_scope_can_be_url(self):
        self.sdk.search_institutions('bank', scope='url')

    def test_search_scope_can_be_name(self):
        self.sdk.search_institutions('bank', scope='name')

    def test_search_defaults_to_first_page(self):
        result = self.sdk.search_institutions('bank')
        assert result.current_page == 1

    def test_search_page_can_be_changed(self):
        page_one = self.sdk.search_institutions('bank')
        result = self.sdk.search_institutions('bank', page=2)
        assert result.current_page == 2
        assert result.data != page_one.data

    def test_submit_key_serialization(self):
        prompt = geezeo.AuthPrompt({
            "id": 2,
            "fi_id": 20349,
            "name": "CashEdge Test Bank (Agg) - Retail Non 2FA",
            "url": "https://cashbank.cashedge.com/.../LoginPage.jsp",
            "ce_login_parameters": [
                {
                    "id": 3,
                    "parameter_id": "42971",
                    "parameter_caption": "UserName",
                    "parameter_type": "login",
                    "parameter_max_length": 20
                },
                {
                    "id": 4,
                    "parameter_id": "42972",
                    "parameter_caption": "Password",
                    "parameter_type": "password",
                    "parameter_max_length": 20
                }
            ]
        })
        assert type(prompt.submit_key) is six.text_type

    def test_submit_key_deserialization(self):
        prompt = geezeo.AuthPrompt({
            "id": 2,
            "fi_id": 20349,
            "name": "CashEdge Test Bank (Agg) - Retail Non 2FA",
            "url": "https://cashbank.cashedge.com/.../LoginPage.jsp",
            "ce_login_parameters": [
                {
                    "id": 3,
                    "parameter_id": "42971",
                    "parameter_caption": "UserName",
                    "parameter_type": "login",
                    "parameter_max_length": 20
                },
                {
                    "id": 4,
                    "parameter_id": "42972",
                    "parameter_caption": "Password",
                    "parameter_type": "password",
                    "parameter_max_length": 20
                }
            ]
        })
        key_data = geezeo.decode_submit_key(prompt.submit_key)
        assert key_data == {
            'type': 'auth',
            'id': '2',
            'name': 'CashEdge Test Bank (Agg) - Retail Non 2FA',
            'parameters': [
                {
                    'key': '42971',
                    'caption': 'UserName',
                    'type': 'login',
                    'max_length': 20,
                },
                {
                    'key': '42972',
                    'caption': 'Password',
                    'type': 'password',
                    'max_length': 20,
                },
            ]
        }

    def test_non_mfa_login(self):
        prompts = self.sdk.search_institutions('CashEdge Test Bank Agg')
        for p in prompts:
            if p.name == "CashEdge Test Bank (Agg) - Retail Non 2FA":
                prompt = p
                break
        auth_info = {}
        for param in prompt.parameters:
            if param.caption == 'UserName':
                auth_info[param.key] = 'script1'
            elif param.caption == 'Password':
                auth_info[param.key] = 'cashedge1'
        result = self.sdk.authenticate(prompt.submit_key, auth_info)
        assert isinstance(result, geezeo.AggregatedInstitution)

    def test_mfa_login(self):
        prompts = self.sdk.search_institutions('CashEdge Test Bank Agg')
        for p in prompts:
            if p.name == "CashEdge Test Bank (Agg) - Retail 2FA":
                prompt = p
                break
        auth_info = {}
        for param in prompt.parameters:
            if param.caption == 'User ID':
                auth_info[param.key] = 'test'
            elif param.caption == 'Password':
                auth_info[param.key] = 'test'
        try:
            self.sdk.authenticate(prompt.submit_key, auth_info)
        except geezeo.MFARequiredError as e:
            assert len(e.auth_prompt.parameters) == 1
            parameters = {e.auth_prompt.parameters[0].key: 'red'}
            self.sdk.authenticate(e.auth_prompt.submit_key, parameters)
        else:
            assert False, "Didn't get MFA prompt."


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools.command.test import test as TestCommand
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    'requests>=2.8.1',
    'six>=1.10.0',
]

test_requirements = [
    'tox>=2.1.1',
]

setup(
    name='geezeo',
    version='0.2.0',
    description="Connect to the Geezeo API.",
    long_description=readme + '\n\n' + history,
    author="Geezeo",
    author_email='ddonna@amstonstudio.com',
    url='https://github.com/Geezeo/sdk-python',
    packages=[
        'geezeo',
    ],
    package_dir={'geezeo':
                 'geezeo'},
    include_package_data=True,
    install_requires=requirements,
    license="",
    zip_safe=False,
    keywords='geezeo',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    cmdclass={'test': Tox},
)



# -*- coding: utf-8 -*-

"""
DTOs and some logic for managing aggregated accounts.
"""

import base64
import json
from . import util


def encode_submit_key(data):
    """
    Turn a JSON-serializable object into a string of echo data.

    :param dict data: arbitrary JSON-serializable object
    :return: string of base 64
    """

    json_data = json.dumps(data)
    json_bytes = json_data.encode('ascii')
    base64_bytes = base64.b64encode(json_bytes)
    submit_key = base64_bytes.decode('ascii')
    return submit_key


def decode_submit_key(submit_key):
    """
    Turn a submit key back into data.

    :param str submit_key: echo data saved from an AuthPrompt object
    :return: dictionary of submit key data
    """

    base64_bytes = submit_key.encode('ascii')
    json_bytes = base64.b64decode(base64_bytes)
    json_data = json_bytes.decode('ascii')
    data = json.loads(json_data)
    return data


class AuthPrompt(object):

    """
    Login form for a user to authenticate with a financial institution.

    Geezeo uses a third-party financial aggregator to provide information on
    accounts and transations for multiple financial institutions. This class
    is essentially a login form for such an institution.

    .. attribute:: name

        The name of the institution with whom this prompt will let the user
        authenticate.

    .. attribute:: id

        Unique string identifier for the institution involved.

    .. attribute:: submit_key

        String that uniquely identifies this prompt. This will need to be
        passed to :meth:`SDK.authenticate` in order to submit the prompt.

    .. attribute:: parameters

        List of :class:`AuthParameter` objects. The set of :attr:`key
        <AuthParameter.key>` attributes indicate what keys need to be present
        in the ``responses`` argument of :meth:`submit`.
    """

    def __init__(self, data):
        # {
        #   "id": 2,
        #   "fi_id": 20349,
        #   "name": "CashEdge Test Bank (Agg) - Retail Non 2FA",
        #   "url": "https://cashbank.cashedge.com/.../LoginPage.jsp",
        #   "ce_login_parameters": [
        #     {
        #       "id": 3,
        #       "parameter_id": "42971",
        #       "parameter_caption": "UserName",
        #       "parameter_type": "login",
        #       "parameter_max_length": 20
        #     },
        #     {
        #       "id": 4,
        #       "parameter_id": "42972",
        #       "parameter_caption": "Password",
        #       "parameter_type": "password",
        #       "parameter_max_length": 20
        #     }
        #   ]
        # }

        self.name = data['name']
        self.id = str(data['id'])
        self.parameters = [AuthParameter(param)
                           for param in data['ce_login_parameters']]

        self.submit_key = encode_submit_key({
            'type': 'auth',
            'id': self.id,
            'name': self.name,
            'parameters': [param.to_json() for param in self.parameters],
        })

    def to_json(self):
        """
        Render as JSON.

        .. code-block:: javascript

            {
                "name": "Horse Financial",
                "id": "4321",
                "submitKey": "eyJvdGhlcl90aGluZyI6ICIyZXJkc2Fwc2EiLCAidGh=",
                "parameters": [
                    {
                        "key": "8765",
                        "caption": "User name",
                        "type": "username",
                        "maxLength": 20
                    },
                    {
                        "key": "8766",
                        "caption": "Password",
                        "type": "password",
                        "maxLength": null
                    }
                ]
            }
        """

        return {
            'name': self.name,
            'id': self.id,
            'parameters': [param.to_json() for param in self.parameters],
            'submitKey': self.submit_key,
        }


class AggregatedInstitution(util.Institution):

    """
    Institution with aggregated accounts.

    .. attribute:: id

        See :attr:`Institution.id`.

    .. attribute:: name

        See :attr:`Institution.name`.

    .. attribute:: submit_key

        Key used to update authentication information for the institution.

    .. attribute:: parameters

        List of :class:`AuthParameter` objects carried over from the
        :class:`AuthPrompt` that was submitted to authenticate with this
        institution.

    .. attribute:: accounts

        List of :class:`Account` objects, one for each account that the user
        holds at the newly authenticated institution.
    """

    def __init__(self, submit_data, accounts):
        self.id = submit_data['id']
        self.name = submit_data['name']

        self.parameters = [AuthParameter.from_json(item)
                           for item in submit_data['parameters']]

        self.submit_key = encode_submit_key({
            'type': 'institution',
            'id': self.id,
            'name': self.name,
            'parameters': submit_data['parameters'],
        })

        self.accounts = accounts

    def to_json(self):
        """
        Render as JSON.

        .. code-block:: javascript

            {
                "id": "4321",
                "name": "Horse Financial",
                "submitKey": "eyJvdGhlcl90aGluZyI6ICIyZXJkc2Fwc2EiLCAidGh=",
                "parameters": [
                    {
                        "key": "8765",
                        "caption": "User name",
                        "type": "username",
                        "maxLength": 20
                    },
                    {
                        "key": "8766",
                        "caption": "Password",
                        "type": "password",
                        "maxLength": null
                    }
                ],
                "accounts": [
                    {
                        "id": "123",
                        "name": "Main Checking",
                        "accountType": "checking"
                    }
                ]
            }
        """

        return {
            'id': self.id,
            'name': self.name,
            'submitKey': self.submit_key,
            'parameters': [p.to_json() for p in self.parameters],
            'accounts': [a.to_json() for a in self.accounts],
        }


class MFAPrompt(AuthPrompt):

    def __init__(self, data, submit_data):
        # {
        #   "mfa_parameters": [
        #     {
        #       "ce_fi_id": null,
        #       "parameter_caption": "What is your favorite color?",
        #       "parameter_id": "answer",
        #       "parameter_max_length": null,
        #       "parameter_type": "password"
        #     }
        #   ],
        #   "session_key": "464a674e7367646...627879413d3d",
        #   "harvest_id": "123770714",
        #   "login_id": "19349692"
        # }

        self.id = submit_data['id']
        self.name = submit_data['name']

        self.parameters = []
        for parameter_data in data['mfa_parameters']:
            self.parameters.append(AuthParameter(parameter_data))

        self.submit_key = encode_submit_key({
            'type': 'mfa',
            'id': self.id,
            'name': self.name,
            'parameters': submit_data['parameters'],

            'session_key': data['session_key'],
            'harvest_id': data['harvest_id'],
            'login_id': data['login_id'],
        })


class MFARequiredError(Exception):

    """
    A financial institution requires multi-factor authentication.

    This class is an :class:`Exception` that can be raised during attempts to
    communicate with an aggregated financial institution, such as when calling
    :meth:`SDK.authenticate`. If an :class:`MFArequiredError` exception is
    raised, the institution has issued a multi-factor auth challenge, and a
    follow-up request is needed before the desired data will be returned.

    To continue the transaction, use the new :attr:`auth_prompt`.

    .. attribute:: message

        String describing the prompt as a whole.

    .. attribute:: auth_prompt

        An :attr:`AuthPrompt` that can be submitted to continue the
        authentication process.
    """

    def __init__(self, message, data, submit_data):
        # {
        #   "response": {
        #     "message": "The account requires further authentication",
        #     "data": {
        #       "mfa_parameters": [
        #         {
        #           "ce_fi_id": null,
        #           "parameter_caption": "What is your favorite color?",
        #           "parameter_id": "answer",
        #           "parameter_max_length": null,
        #           "parameter_type": "password"
        #         }
        #       ],
        #       "session_key": "464a674e7367646...627879413d3d",
        #       "harvest_id": "123770714",
        #       "login_id": "19349692"
        #     }
        #   }
        # }

        self.message = message
        self.auth_prompt = MFAPrompt(data, submit_data)

    def to_json(self):
        data = self.auth_prompt.to_json
        data['message'] = self.message
        return data


class AuthParameter(object):
    """
    One part of a multi-factor authentication prompt.

    .. attribute:: key

        String that identifies this parameter uniquely among the parameters
        for the parent prompt.

    .. attribute:: caption

        String challenge for the user, e.g. ``"What is your favorite
        color?"``.

    .. attribute:: type

        String description of the information to be provided, e.g.
        ``"password"``.

    .. attribute:: max_length

        Maximum length that the answer for the parameter can have (int).
        This may be ``None``, indicating that there is no max length.
    """

    def __init__(self, data):
        # {
        #   "id": 3,
        #   "parameter_id": "42971",
        #   "parameter_caption": "UserName",
        #   "parameter_type": "login",
        #   "parameter_max_length": 20
        # }

        self.key = data['parameter_id']
        self.caption = data['parameter_caption']
        self.type = data['parameter_type']
        self.max_length = data['parameter_max_length']

    @classmethod
    def from_json(cls, data):
        return cls({
            'parameter_id': data['key'],
            'parameter_caption': data['caption'],
            'parameter_type': data['type'],
            'parameter_max_length': data['maxLength'],
        })

    def to_json(self):
        """
        Render as JSON.

        .. code-block:: javascript

            {
                "key": "8765",
                "caption": "User name",
                "type": "username",
                "maxLength": 20
            }
        """

        return {
            'key': self.key,
            'caption': self.caption,
            'type': self.type,
            'maxLength': self.max_length,
        }
